// CsFormatter.h

#ifndef _CS_FORMATTER_H_
#define _CS_FORMATTER_H_


#include "CsParserTypes.h"
#include "FastDeque.h"
#include "CharDeque.h"

//------------------------------------------------------------------------------
// CsFormatTokenType
//------------------------------------------------------------------------------

enum CsFormatTokenType
{
	tkF_WS = TOKEN_COUNT,
	tkF_WS_KILL,
	tkF_NEWLINE,
	tkF_NEWLINE_KILL,
	tkF_INC_TAB,
	tkF_DEC_TAB,
	tkF_PP_INC_TAB,
	tkF_PP_DEC_TAB,
};

//------------------------------------------------------------------------------
// CsFormatTokenPriority
//------------------------------------------------------------------------------

//enum CsFormatTokenPriority
//{
//	ftp_default = 0, // 
//	ftp_tab_or_newline = 1, // tab or newline kill whitespaces
//};

//------------------------------------------------------------------------------
// CsFormatter
//------------------------------------------------------------------------------

class CsFormatter : public CsAstCallback
{
	enum generation_option
	{
		go_generate_tokens,
		go_generate_text,
	};

public:
	CsFormatter();
	~CsFormatter();
	void setDefault();
	void generate ( CsNode * node );
	void generate_text();
	void generate_tokens();

// callback
	virtual void callback ( CsCompoundIdentifier & compound_identifier );
	virtual void callback ( CsModifiers & node );
	virtual void callback ( CsTypeSpecifier & node );
	virtual void callback ( CsTypeSpecifiers * node );

	//virtual void callback ( CsUsingDirective * node );
	virtual void callback ( CsUsingAliasDirective * node );
	virtual void callback ( CsUsingNamespaceDirective * node );
	virtual void callback ( CsTypeRef * node );
	virtual void callback ( CsSimpleName * node );
	virtual void callback ( CsLiteral * node );
	virtual void callback ( CsEmbeddedTypeMemberAccess * node );
	virtual void callback ( CsThisAccess * node );
	virtual void callback ( CsBaseMemberAccess * node );
	virtual void callback ( CsBaseIndexerAccess * node );
	virtual void callback ( CsTypeofExpression * node );
	virtual void callback ( CsCheckedExpression * node );
	virtual void callback ( CsUncheckedExpression * node );
	virtual void callback ( CsLockStatement * node );
	virtual void callback ( CsSizeofExpression * node );
	virtual void callback ( CsParenthesizedExpression * node );
	virtual void callback ( CsNewObjectExpression * node );
	virtual void callback ( CsNewArrayExpression * node );
	virtual void callback ( CsInvocationExpression * node );
	virtual void callback ( CsElementAccess * node );
	virtual void callback ( CsMemberAccess * node );
	virtual void callback ( CsPointerMemberAccess * node );
	virtual void callback ( CsPostIncrementDecrementExpression * node );
	virtual void callback ( CsBinaryExpression * node );
	virtual void callback ( CsAsIsExpression * node );
	virtual void callback ( CsConditionalExpression * node );
	virtual void callback ( CsOperatorUnaryExpression * node );
	virtual void callback ( CsPreIncrementDecrementExpression * node );
	virtual void callback ( CsCastUnaryExpression * node );
	virtual void callback ( CsAssignmentExpression * node );
	virtual void callback ( CsExpressionList * node );
	virtual void callback ( CsNamedArgument * node );
	virtual void callback ( CsAttribute * node );
	virtual void callback ( CsAttributeSection * node );
	virtual void callback ( CsAttributes * node );
	virtual void callback ( CsTypeBase * node );
	virtual void callback ( CsClass * node );
	virtual void callback ( CsStruct * node );
	virtual void callback ( CsInterface * node );
	virtual void callback ( CsEnumMember * node );
	virtual void callback ( CsEnum * node );
	virtual void callback ( CsFormalParameter * node );
	virtual void callback ( CsFormalParameterList * node );
	virtual void callback ( CsDelegate * node );
	virtual void callback ( CsNamespace * node );
	virtual void callback ( CsArgument * node );
	virtual void callback ( CsArgumentList * node );
	virtual void callback ( CsBlock * node );
	virtual void callback ( CsLabel * node );
	virtual void callback ( CsLabeledStatement * node );
	virtual void callback ( CsEmptyStatement * node );
	virtual void callback ( CsExpressionStatement * node );
	virtual void callback ( CsDeclarationStatement * node );
	virtual void callback ( CsLocalConstantDeclarator * node );
	virtual void callback ( CsLocalConstantDeclaration * node );
	virtual void callback ( CsIfStatement * node );
	virtual void callback ( CsSwitchLabel * node );
	virtual void callback ( CsSwitchSection * node );
	virtual void callback ( CsSwitchStatement * node );
	virtual void callback ( CsWhileStatement * node );
	virtual void callback ( CsDoWhileStatement * node );
	virtual void callback ( CsLocalVariableDeclarator * node );
	virtual void callback ( CsLocalVariableDeclaration * node );
	virtual void callback ( CsStatementExpressionList * node );
	virtual void callback ( CsForStatement * node );
	virtual void callback ( CsForeachStatement * node );
	virtual void callback ( CsBreakStatement * node );
	virtual void callback ( CsContinueStatement * node );
	virtual void callback ( CsGotoStatement * node );
	virtual void callback ( CsReturnStatement * node );
	virtual void callback ( CsThrowStatement * node );
	virtual void callback ( CsCatchClause * node );
	virtual void callback ( CsTryStatement * node );
	virtual void callback ( CsCheckedStatement * node );
	virtual void callback ( CsUncheckedStatement * node );
	virtual void callback ( CsUsingStatement * node );
	virtual void callback ( CsUnsafeStatement * node );
	virtual void callback ( CsFixedPointDeclarator * node );
	virtual void callback ( CsFixedStatement * node );
	virtual void callback ( CsStackAllocExpression * node );
	virtual void callback ( CsConstructorDeclaration * node );
	virtual void callback ( CsConstantDeclarator * node );
	virtual void callback ( CsConstantDeclaration * node );
	virtual void callback ( CsVariableDeclarator * node );
	virtual void callback ( CsVariableDeclaration * node );
	virtual void callback ( CsArrayInitializer * node );
	virtual void callback ( CsMethod * node );
	virtual void callback ( CsPropertyAccessor * node );
	virtual void callback ( CsProperty * node );
	virtual void callback ( CsEventAccessor * node );
	virtual void callback ( CsEventDeclarator * node );
	virtual void callback ( CsEvent * node );
	virtual void callback ( CsIndexer * node );
	virtual void callback ( CsOperator * node );
	virtual void callback ( CsConversionOperatorDeclaration * node );
	virtual void callback ( CsDestructorDeclaration * node );
	virtual void callback ( CsCompilationUnit * node );
	virtual void callback ( CsMakeRefExpression * node );
	virtual void callback ( CsRefTypeExpression * node );
	virtual void callback ( CsRefValueExpression * node );
	virtual void callback ( CsArgListExpression * node );

	inline void addTokens ( const FastArray<CsToken> & tokens )
	{
		tokens_.push_back ( tokens.begin(), tokens.end() );
	}

	inline void addToken ( CsToken token )
	{
		tokens_.push_back ( token );
	}

	inline void addToken ( WPHashString phs )
	{
		if ( phs )
			tokens_.push_back ( CsToken ( tkIDENTIFIER, (PHashString) phs, LineInfo() ) );
	}

	inline void addToken ( CsTokenType token_type )
	{
		tokens_.push_back ( CsToken ( token_type, 0, LineInfo() ) );
	}

	inline void addKeyword ( CsTokenType token_type )
	{
		addTokens ( keyword_around );
		tokens_.push_back ( CsToken ( token_type, 0, LineInfo() ) );
		addTokens ( keyword_around );
	}

	inline void addOperator ( CsTokenType token_type )
	{
		addTokens ( operator_around );
		tokens_.push_back ( CsToken ( token_type, 0, LineInfo() ) );
		addTokens ( operator_around );
	}

	//static size_t st_index;

	inline void addWS()
	{
//#ifdef _DEBUG
//		size_t index = std::distance ( tokens_.begin(), tokens_.end() );
//		if ( index >= st_index )
//		{
//			XASSERT ( 0 );
//		}
//#endif // _DEBUG
		tokens_.push_back ( CsToken ( CsTokenType ( tkF_WS ), 1, LineInfo() ) );
	}

	inline void addWSKill()
	{
		tokens_.push_back ( CsToken ( CsTokenType ( tkF_WS_KILL ), 0, LineInfo() ) );
	}

	inline void addNewline()
	{
		addWSKill();
		tokens_.push_back ( CsToken ( CsTokenType ( tkF_NEWLINE ), 1, LineInfo() ) );
		addWSKill();
	}

	inline void addNewline2()
	{
		addWSKill();
		tokens_.push_back ( CsToken ( CsTokenType ( tkF_NEWLINE ), 2, LineInfo() ) );
		addWSKill();
	}

	inline void addNewlineKill0()
	{
		addWSKill();
		tokens_.push_back ( CsToken ( CsTokenType ( tkF_NEWLINE_KILL ), 0, LineInfo() ) );
		addWSKill();
	}

	inline void addNewlineKill1()
	{
		addWSKill();
		tokens_.push_back ( CsToken ( CsTokenType ( tkF_NEWLINE_KILL ), 1, LineInfo() ) );
		addWSKill();
	}

	inline void addComma()
	{
		addWSKill();
		tokens_.push_back ( CsToken ( CsTokenType ( tkCOMMA ), 1, LineInfo() ) );
		tokens_.push_back ( CsToken ( CsTokenType ( tkF_WS ), 1, LineInfo() ) );
	}

	inline void addColon()
	{
		addWSKill();
		tokens_.push_back ( CsToken ( CsTokenType ( tkCOLON ), 1, LineInfo() ) );
	}

	inline void addSemicolon()
	{
		addWSKill();
		tokens_.push_back ( CsToken ( CsTokenType ( tkSEMICOLON ), 1, LineInfo() ) );
	}

	inline void addIncTab()
	{
		tokens_.push_back ( CsToken ( CsTokenType ( tkF_INC_TAB ), 1, LineInfo() ) );
	}

	inline void addDecTab()
	{
		tokens_.push_back ( CsToken ( CsTokenType ( tkF_DEC_TAB ), 1, LineInfo() ) );
	}

// tokens
	MediumFastDeque<CsToken> tokens_;
	CharDeque generated_text_;
	MediumFastDeque<CsToken> generated_tokens_;
	unsigned n_tab;

// CsCompoundIdentifier
	FastArray<CsToken> compound_identifier_before_after;
	FastArray<CsToken> compound_identifier_around_dot;
	FastArray<CsToken> keyword_around;
	FastArray<CsToken> operator_around;

// options
	LineEnum end_of_line_type_;
	bool storeNewlines_;
	generation_option generation_option_;
	static const size_t tab_token_buffer_size_start_ = 8;
	size_t tab_token_buffer_size_;
	char * tab_token_buffer_;
	static const size_t ws_token_buffer_size_start_ = 8;
	size_t ws_token_buffer_size_;
	char * ws_token_buffer_;
};

//------------------------------------------------------------------------------


#endif // _CS_FORMATTER_H_
